import datetime
import logging
import uuid
from enum import Enum
from typing import Optional

from django.core import exceptions
from django.db import transaction
from django.http.response import JsonResponse
from django.utils import timezone
from django.utils.timezone import now
from rest_framework import status
from rest_framework.authentication import SessionAuthentication, BasicAuthentication
from rest_framework.generics import CreateAPIView, GenericAPIView
from rest_framework.mixins import UpdateModelMixin
from rest_framework.generics import CreateAPIView, GenericAPIView, DestroyAPIView
from rest_framework.parsers import JSONParser
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.views import APIView

from greg.models import Identity, InvitationLink, Person, Invitation
from greg.permissions import IsSponsor
from gregui.api.serializers.guest import GuestRegisterSerializer
from gregui.api.serializers.invitation import InviteGuestSerializer
from gregui.mailutils import send_invite_mail
from gregui.models import GregUserProfile

logger = logging.getLogger(__name__)


class InvitationView(CreateAPIView, DestroyAPIView):
    """
    Endpoint for invitation creation and cancelling


    {
        "first_name": "dfff",
        "last_name": "sss",
        "date_of_birth": null,
        "role": {
            "orgunit": 1,
            "start_date": null,
            "type": 1,
            "end_date": "2021-12-15",
            "contact_person_unit": "",
            "comments": "",
            "available_in_search": false
        }
    }
    """

    authentication_classes = [BasicAuthentication, SessionAuthentication]
    permission_classes = [IsSponsor]
    parser_classes = [JSONParser]
    serializer_class = InviteGuestSerializer

    def post(self, request, *args, **kwargs) -> Response:
        """
        Invitation creation endpoint

        Restricted to Sponsors, and a sponsor can only invite guests to OUs
        they are registered to.

        The next step in the flow is for the guest to use their link, review the
        information, and confirm it.
        """
        sponsor_user = GregUserProfile.objects.get(user=request.user)
        serializer = self.serializer_class(
            data=request.data,
            context={"request": request, "sponsor": sponsor_user.sponsor},
        )
        serializer.is_valid(raise_exception=True)
        person = serializer.save()

        for invitationlink in InvitationLink.objects.filter(
            invitation__role__person_id=person.id,
            invitation__role__sponsor_id=sponsor_user.sponsor_id,
        ):
            send_invite_mail(invitationlink)

        return Response(status=status.HTTP_201_CREATED)

    def delete(self, request, *args, **kwargs) -> Response:
        person_id = int(request.query_params["person_id"])

        # TODO Determine if person should be deleted as well
        person = Person.objects.get(id=person_id)
        if person.is_registered or person.is_verified:
            # The guest has already gone through the registration step. The guest should
            # not be verified, but including that check just in case here.
            # If this is the case then there is an unexpected situation, the cancel option
            # should only apply to guests that have not completed the registration
            logger.warning(
                f"Attempting to delete invitation for already registered guest with person ID {person_id}"
            )
            return Response(status.HTTP_400_BAD_REQUEST)

        # Since the person is not registered, all roles attached to invitations for him should be removed
        invitationlinks = InvitationLink.objects.filter(
            invitation__role__person_id=person_id
        )
        roles = set(
            [invitationlink.invitation.role for invitationlink in invitationlinks]
        )

        for role in roles:
            # Delete the role, the cascading will cause all the invitation links connected
            # to it to be removed as well
            role.delete()

        return Response(status=status.HTTP_200_OK)


class CheckInvitationView(APIView):
    authentication_classes = []
    permission_classes = [AllowAny]

    def post(self, request, *args, **kwargs):
        """
        Endpoint for verifying and setting invite_id in session.

        This endpoint is meant to be called in the background by the frontend on the
        page you get to by following the invitation url to the frontend. This way a
        session is created keeping the invite id safe, until the user returns from
        feide login if they choose to use it.

        Uses post to prevent invite id from showing up in various logs.
        """
        invite_id = request.data.get("uuid")
        if not invite_id:
            return Response(status=status.HTTP_403_FORBIDDEN)
        try:
            invite_link = InvitationLink.objects.get(uuid=invite_id)
        except (InvitationLink.DoesNotExist, exceptions.ValidationError):
            return Response(status=status.HTTP_403_FORBIDDEN)
        if invite_link.expire <= timezone.now():
            return Response(status=status.HTTP_403_FORBIDDEN)
        request.session["invite_id"] = invite_id
        return Response(status=status.HTTP_200_OK)


class SessionType(Enum):
    INVITE = "invite"
    FEIDE = "feide"


class InvitedGuestView(GenericAPIView):
    authentication_classes = [SessionAuthentication, BasicAuthentication]
    # The endpoint is only for invited guests, but the authorization happens in the actual method
    permission_classes = [AllowAny]
    parser_classes = [JSONParser]
    serializer_class = GuestRegisterSerializer

    # TODO Update to make dynamic based on where we get the information from. If we get some from Feide, then the user should not be allowed to change it
    fields_allowed_to_update = ["email", "fnr", "mobile_phone", "passport"]

    def get(self, request, *args, **kwargs):
        """
        Endpoint for fetching data related to an invite

        Used by the frontend for fetching data from the backend for review by a guest
        in the frontend, before calling the post endpoint defined below with updated
        info and confirmation.
        """
        invite_id = request.session.get("invite_id")
        try:
            invite_link = InvitationLink.objects.get(uuid=invite_id)
        except (InvitationLink.DoesNotExist, exceptions.ValidationError):
            return Response(status=status.HTTP_403_FORBIDDEN)
        if invite_link.expire <= timezone.now():
            return Response(status=status.HTTP_403_FORBIDDEN)

        # if invite_id:
        invite_link = InvitationLink.objects.get(uuid=invite_id)
        role = invite_link.invitation.role
        person = role.person
        sponsor = role.sponsor

        # If the user is not logged in then tell the client to take him through the manual registration process
        session_type = (
            SessionType.INVITE.value
            if request.user.is_anonymous
            else SessionType.FEIDE.value
        )

        fnr = self._get_identity_or_none(
            person, Identity.IdentityType.NORWEGIAN_NATIONAL_ID_NUMBER
        )
        passport = self._get_identity_or_none(
            person, Identity.IdentityType.PASSPORT_NUMBER
        )
        feide_id = self._get_identity_or_none(person, Identity.IdentityType.FEIDE_ID)

        data = {
            "person": {
                "first_name": person.first_name,
                "last_name": person.last_name,
                "email": person.private_email and person.private_email.value,
                "mobile_phone": person.private_mobile and person.private_mobile.value,
                "fnr": fnr,
                "passport": passport,
                "feide_id": feide_id,
            },
            "sponsor": {
                "first_name": sponsor.first_name,
                "last_name": sponsor.last_name,
            },
            "role": {
                "ou_name_nb": role.orgunit.name_nb,
                "ou_name_en": role.orgunit.name_en,
                "role_name_nb": role.type.name_nb,
                "role_name_en": role.type.name_en,
                "start": role.start_date,
                "end": role.end_date,
                "comments": role.comments,
            },
            "meta": {"session_type": session_type},
        }
        return JsonResponse(data=data, status=status.HTTP_200_OK)

    def post(self, request, *args, **kwargs):
        """
        Endpoint for confirmation of data updated by guest

        Used by frontend when the confirm button at the end of the review flow is
        clicked by the user. The next part of the flow is for the sponsor to confirm
        the guest.
        """
        invite_id = request.session.get("invite_id")

        # Ensure the invitation link is valid and not expired
        try:
            invite_link = InvitationLink.objects.get(uuid=invite_id)
        except (InvitationLink.DoesNotExist, exceptions.ValidationError):
            return Response(status=status.HTTP_403_FORBIDDEN)
        if invite_link.expire <= timezone.now():
            return Response(status=status.HTTP_403_FORBIDDEN)

        person = invite_link.invitation.role.person

        data = request.data

        if self._verified_fnr_already_exists(person) and "fnr" in data:
            # The user should not be allowed to change a verified fnr
            return Response(status=status.HTTP_400_BAD_REQUEST)

        if not self._only_allowed_fields_in_request(data):
            return Response(status=status.HTTP_400_BAD_REQUEST)

        with transaction.atomic():
            # Note this only serializes data for the person, it does not look at other sections
            # in the response that happen to be there
            serializer = self.get_serializer(
                instance=person, data=request.data["person"]
            )
            serializer.is_valid(raise_exception=True)
            person = serializer.save()

            # Mark guest interaction done
            person.registration_completed_date = timezone.now().date()
            person.save()

            # Expire the invite link
            invite_link.expire = timezone.now()
            invite_link.save()
            # TODO: Send an email to the sponsor?
        return Response(status=status.HTTP_200_OK)

    @staticmethod
    def _verified_fnr_already_exists(person) -> bool:
        try:
            person.identities.get(
                type=Identity.IdentityType.NORWEGIAN_NATIONAL_ID_NUMBER,
                verified=Identity.Verified.AUTOMATIC,
            )
            return True
        except Identity.DoesNotExist:
            return False

    def _only_allowed_fields_in_request(self, request_data) -> bool:
        # Check how many of the allowed fields are filled in
        person_data = request_data["person"]
        number_of_fields_filled_in = sum(
            map(lambda x: x in person_data.keys(), self.fields_allowed_to_update)
        )
        # Check that there are no other fields filled in
        return number_of_fields_filled_in == len(person_data.keys())

    @staticmethod
    def _get_identity_or_none(
        person: Person, identity_type: Identity.IdentityType
    ) -> Optional[str]:
        try:
            return person.identities.get(type=identity_type).value
        except Identity.DoesNotExist:
            return None


class ResendInvitationView(UpdateModelMixin, APIView):
    authentication_classes = [BasicAuthentication, SessionAuthentication]
    permission_classes = [IsSponsor]

    def patch(self, request, *args, **kwargs) -> Response:
        person_id = kwargs["person_id"]

        invitation_links = InvitationLink.objects.filter(
            invitation__role__person__id=person_id
        )

        if invitation_links.count() == 0:
            # No invitation, not expected that the endpoint should be called in this case
            return Response(status=status.HTTP_400_BAD_REQUEST)

        if invitation_links.count() > 1:
            logger.warning(f"Person with ID {person_id} has multiple invitation links")

        for link in invitation_links:
            link.uuid = uuid.uuid4()
            link.expire = timezone.now() + datetime.timedelta(days=30)
            link.save()
            # invitationlink = InvitationLink.objects.create(
            #     invitation=link.invitation,
            #     expire=timezone.now() + datetime.timedelta(days=30),
            # )

            send_invite_mail(link)

        return Response(status=status.HTTP_200_OK)
